// Splay-   


#include<stdio.h>
//#include<memory>
#include<algorithm>
//#include<assert.h>


class SplayTree
{
	//!    .
	//!  ,    
	//! LR -     ,  .
	enum State{
		undef=0,			//!<  
		L=1,
		R=2,
		LL=(L<<2)+L, 
		LR=(L<<2)+R,
		RL=(R<<2)+L,
		RR=(R<<2)+R,
		idle=L|R,			//!<  
	};
protected:
	//!    
	typedef unsigned int value_type;
	//!  
	struct Node{
		value_type val;
		Node *p1, *p2;
		Node(value_type val) :val(val), p1(0), p2(0) {}
	};
	//!      
	struct TreePosition{
		Node *node;
		Node *parent;
	};
public:

	SplayTree() :treeRoot(0) {}
	~SplayTree() { Clear(); }

	bool Insert(value_type);			//!< 
	bool Remove(value_type);			//!< 
	bool FindPassive(value_type) const;	//!< 
	bool FindAdjust (value_type);		//!<   
protected:
	Node *treeRoot;
	value_type GetMax(Node*&);			//!<    
	bool PickUp(value_type val);		//!<     
private:
	void  Clear();
	// rotate & state
	void  RotateL(Node*&) const;		//!< 
	void  RotateR(Node*&) const;
	State StateL (State ) const;		//!<    
	State StateR (State ) const;
	// move
	void  MoveDownL(TreePosition&, value_type val);		//!<    
	void  MoveDownR(TreePosition&, value_type val);		//!<
	State MoveUp   (TreePosition&);						//!<    
	// move inside
	void  SwapL  (Node*&, Node*&, Node*&) const;		//!<    
	void  MoveUpL(TreePosition&);
	State MoveFix(Node*);								//!<     

	// tree walk to path
	void WalkDown(TreePosition&, value_type val);		//!<      
	bool WalkUp  (TreePosition&, State);				//!<      (   Splay )
	// splay
	State SplayOne (Node*&,State);						//!<  
	void  SplayRoot(Node*&,State);						//!<   
};


void SplayTree::Clear()
{
	while (treeRoot) {
		while (treeRoot->p1) RotateR(treeRoot);
		std::auto_ptr<Node> ref(treeRoot);
		treeRoot = treeRoot->p2;
	}
}

void SplayTree::RotateL(Node *&node) const
{
	SwapL(node, node->p2, node->p2->p1);
}

void SplayTree::RotateR(Node *&node) const
{
	SwapL(node, node->p1, node->p1->p2);
}

SplayTree::State SplayTree::StateL(State st) const
{
	return st == idle ? st : State(L | st << 2);
}

SplayTree::State SplayTree::StateR(State st) const
{
	return st == idle ? st : State(R | st << 2);
}

void SplayTree::SwapL(Node *&n1, Node *&n2, Node *&n3) const
{
	Node *t = n1;
	n1 = n2;
	n2 = n3;
	n3 = t;
}

void SplayTree::MoveDownL(TreePosition &it, value_type val)
{
	SwapL(it.node->p1, it.parent, it.node);
}

void SplayTree::MoveDownR(TreePosition &it, value_type val)
{
	std::swap(it.node->p1, it.node->p2);
	MoveDownL(it, val);
}

SplayTree::State SplayTree::MoveFix(Node *node)
{
	if (
		(!node->p1 || node->p1->val <  node->val) &&
		(!node->p2 || node->val <= node->p2->val)
		)
		return L;
	std::swap(node->p1, node->p2);
	return R;
}

void SplayTree::MoveUpL(TreePosition &it)
{
	SwapL(it.node,it.parent,it.parent->p1);
}


SplayTree::State SplayTree::MoveUp(TreePosition &it)
{
	MoveUpL(it);
	return MoveFix(it.node);
}

void SplayTree::WalkDown(TreePosition &it, value_type val)
{
	it.parent = 0;
	it.node   = treeRoot;
	while (it.node)
	{
		if (val == it.node->val) break;
		(val < it.node->val) ? MoveDownL(it, val) : MoveDownR(it, val);
	}
}

bool SplayTree::WalkUp(TreePosition &it, State st)
{
	while (it.parent)
	{
		switch (MoveUp(it))
		{
		case L:  st = StateL(SplayOne(it.node->p1, st)); break;
		case R:  st = StateR(SplayOne(it.node->p2, st)); break;
		}
	}
	SplayOne (treeRoot, st);
	SplayRoot(treeRoot, st);
	return st==idle;
}

void SplayTree::SplayRoot(Node *&node, State st)
{
	switch (st)
	{
	case L: RotateR(node); break;
	case R: RotateL(node); break;
	}
}


SplayTree::State SplayTree::SplayOne(Node *&node, State st)
{
	switch (st)
	{
	case LL: RotateR(node    ); break;
	case RR: RotateL(node    ); break;
	case RL: RotateL(node->p1); break;
	case LR: RotateR(node->p2); break;
	default: return st;
	}
	SplayRoot(node, State(st &(R|L)));
	return undef;
}

SplayTree::value_type SplayTree::GetMax(Node *&node)
{
	Node **ref = &node;
	while ((*ref)->p2) ref = &(*ref)->p2;
	std::auto_ptr<Node> obj(*ref);
	*ref = (*ref)->p1;
	return obj->val;
}


//! \result true    
bool SplayTree::PickUp(value_type val)
{
	TreePosition it;
	WalkDown(it, val);
	return WalkUp(it, it.node ? undef : idle);
}

//! \result true    
bool SplayTree::Insert(value_type val)
{
	if (!treeRoot)
	{
		treeRoot = new Node(val);
		return false;
	}
	TreePosition it;
	WalkDown(it, val);
	if (!it.node) it.node = new Node(val);
	return !WalkUp(it, it.node ? idle : undef);
}


//! \result true    
bool SplayTree::Remove(value_type val)
{
	if (PickUp(val)) return true;
	if (treeRoot->p1) treeRoot->val = GetMax(treeRoot->p1);
	else{
		std::auto_ptr<Node> ref(treeRoot);
		treeRoot = treeRoot->p2;
	}
	return false;
}


//! \result true    
bool SplayTree::FindPassive(value_type val) const
{
	for (Node *node = treeRoot; node;)
	{
		if (val == node->val) return false;
		node = (val < node->val) ? node->p1 : node->p2;
	}
	return true;
}

//! \result true    
bool SplayTree::FindAdjust(value_type val)
{
	return !PickUp(val);
}




int main()
{
	SplayTree tree;
	for (int p = 0; p<1000; p++)
	{
		tree.Insert(rand() % 100);
		tree.Remove(rand() % 100);
	}
	return 0;
}

